// app global css in SCSS form

@import '../../node_modules/xbbcode-parser/xbbcode.css';

.border-hover {
  border-radius: 4px;
  border: 1px solid transparent;
}

.border-hover:hover {
  border: 1px solid rgba(0, 0, 0, 0.24);
  transition: border-color 0.36s cubic-bezier(0.4, 0, 0.2, 1);
}

.header-background {
  background: $header !important
}
.footer-background {
  background: $header !important
}
.bg-card-head {
  background: var(--q-card-head) !important
}
